﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Visualization;

using Rectangle = HeuristicLab.Visualization.Rectangle;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views {
  public class SymbolicExpressionTreeTile : Group {
    public double PreferredNodeWidth { get; set; }
    public double PreferredNodeHeight { get; set; }
    private ISymbolicExpressionTree symbolicExpressionTree;
    public ISymbolicExpressionTree SymbolicExpressionTree {
      get { return symbolicExpressionTree; }
      set {
        symbolicExpressionTree = value;
        GeneratePrimitives(PreferredNodeWidth, PreferredNodeHeight);
      }
    }

    public ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode> LayoutEngine { get; set; }
    private Dictionary<IPrimitive, ISymbolicExpressionTreeNode> primitiveMap;

    public SymbolicExpressionTreeTile(IChart chart)
      : base(chart) {
      primitiveMap = new Dictionary<IPrimitive, ISymbolicExpressionTreeNode>();
    }
    public SymbolicExpressionTreeTile(IChart chart, ISymbolicExpressionTree tree)
      : this(chart) {
      SymbolicExpressionTree = tree;
      PreferredNodeWidth = 80;
      PreferredNodeHeight = 40;
    }
    private void GeneratePrimitives(double preferredNodeWidth, double preferredNodeHeight) {
      Clear();
      ISymbolicExpressionTreeNode root = SymbolicExpressionTree.Root;
      if (root.Symbol is ProgramRootSymbol && root.SubtreeCount == 1) { root = root.GetSubtree(0); }
      var visualNodes = LayoutEngine.CalculateLayout(root);

      var font = new Font(FontFamily.GenericSansSerif, 10, GraphicsUnit.Pixel);

      var visualNodeMap = visualNodes.ToDictionary(x => x.Content, x => x);

      foreach (var visualNode in visualNodes) {
        var lowerLeft = new PointD(visualNode.X, visualNode.Y);
        var upperRight = new PointD(visualNode.X + preferredNodeWidth, visualNode.Y + preferredNodeHeight);
        var node = visualNode.Content;
        RectangularPrimitiveBase rectangularPrimitive;
        if (node.SubtreeCount == 0) {
          rectangularPrimitive = new Rectangle(Chart, lowerLeft, upperRight) { Font = font, Text = visualNode.Content.ToString() };
        } else {
          rectangularPrimitive = new Ellipse(Chart, lowerLeft, upperRight) { Font = font, Text = visualNode.Content.ToString() };
        }

        this.Add(rectangularPrimitive);
      }

      foreach (var node in visualNodes.Where(n => n.Content.SubtreeCount > 0)) {
        var parent = node.Content;
        foreach (var child in parent.Subtrees.Select(x => visualNodeMap[x])) {
          var start = new PointD(node.X + preferredNodeWidth / 2, node.Y + preferredNodeHeight);
          var end = new PointD(child.X + preferredNodeWidth / 2, child.Y);
          var line = new Line(this.Chart, start, end);
          this.Add(line);
        }
      }
    }
  }
}
